import { InvalidValueError, MissingEntityError, MissingPropertyError, } from '@crosslab/service-common';
import { logger } from '@crosslab/service-common';
import { randomUUID } from 'crypto';
import { clients } from '../clients/index.js';
import { config } from '../config.js';
import { experimentUrlFromId, getUrlOrInstanceUrl } from './url.js';
export async function buildConnectionPlan(experiment, resolvedDevices) {
    const experimentUrl = experimentUrlFromId(experiment.uuid);
    logger.log('info', 'Building connection plan', { data: { experimentUrl } });
    if (!experiment.devices || experiment.devices.length === 0)
        throw new MissingPropertyError('Experiment must have a device to be run', 400);
    if (!experiment.serviceConfigurations)
        throw new MissingPropertyError('Experiment must have a configuration to be run', 400);
    const pairwiseServiceConfigurations = toPairwiseServiceConfig(experiment.serviceConfigurations);
    const deviceMappedServiceConfigs = mapRoleConfigToDevices(pairwiseServiceConfigurations, experiment);
    const sortedDeviceMappedServiceConfigs = deviceMappedServiceConfigs.map(dmsc => sortServiceParticipantsByDeviceId(dmsc));
    const peerconnections = {};
    for (const serviceConfig of sortedDeviceMappedServiceConfigs) {
        // HOTFIX: for local services: Don't connect local services to each other
        // TODO: create a new connection type 'local' as opposed to 'webrtc' and handle it correctly
        if (getUrlOrInstanceUrl(serviceConfig.devices[0]) ===
            getUrlOrInstanceUrl(serviceConfig.devices[1])) {
            continue;
        }
        const lookupKey = `${getUrlOrInstanceUrl(serviceConfig.devices[0])}::${getUrlOrInstanceUrl(serviceConfig.devices[1])}`;
        if (!(lookupKey in peerconnections)) {
            const deviceA = getResolvedDevice(resolvedDevices, getUrlOrInstanceUrl(serviceConfig.devices[0]));
            const deviceB = getResolvedDevice(resolvedDevices, getUrlOrInstanceUrl(serviceConfig.devices[1]));
            const connectionType = intersection(getSupportedConnectionTypes(deviceA, serviceConfig.serviceType, serviceConfig.participants[0].serviceId), getSupportedConnectionTypes(deviceB, serviceConfig.serviceType, serviceConfig.participants[1].serviceId))[0];
            if (!connectionType)
                throw new InvalidValueError(
                // prettier-ignore
                `Service "${serviceConfig.participants[0].serviceId}:${serviceConfig.participants[1].serviceId}" of type "${serviceConfig.serviceType}" between device "${deviceA.url}" and device "${deviceB.url}" cannot be connected since they have no common supported connection type!`, 400);
            peerconnections[lookupKey] = await createPeerconnection(connectionType, serviceConfig);
        }
        const peerconnection = peerconnections[lookupKey];
        updateServiceConfig(peerconnection.devices[0], getResolvedDevice(resolvedDevices, peerconnection.devices[1].url), serviceConfig, serviceConfig.participants[0], serviceConfig.participants[1]);
        updateServiceConfig(peerconnection.devices[1], getResolvedDevice(resolvedDevices, peerconnection.devices[0].url), serviceConfig, serviceConfig.participants[1], serviceConfig.participants[0]);
    }
    const peerconnectionsArray = Object.values(peerconnections);
    logger.log('info', 'Built connection plan', {
        data: { connectionPlan: peerconnectionsArray },
    });
    return peerconnectionsArray;
}
function sortServiceParticipantsByDeviceId(serviceConfig) {
    const swap = serviceConfig.devices[0].uuid > serviceConfig.devices[1].uuid;
    return {
        ...serviceConfig,
        participants: swap
            ? [serviceConfig.participants[1], serviceConfig.participants[0]]
            : [serviceConfig.participants[0], serviceConfig.participants[1]],
        devices: swap
            ? [serviceConfig.devices[1], serviceConfig.devices[0]]
            : [serviceConfig.devices[0], serviceConfig.devices[1]],
    };
}
function updateServiceConfig(device, remoteDevice, serviceConfig, participant, remoteParticipant) {
    const remoteServiceDescription = remoteDevice.services?.find(service => service.serviceId === remoteParticipant.serviceId &&
        service.serviceType === serviceConfig.serviceType);
    if (!remoteServiceDescription)
        throw new InvalidValueError(`Could not find service description for service "${remoteParticipant.serviceId}" of device "${remoteDevice.url}"`, 400);
    device.config = device.config ?? {};
    device.config.services = device.config.services ?? [];
    device.config?.services?.push({
        ...serviceConfig.configuration,
        ...participant.config,
        serviceId: participant.serviceId,
        serviceType: serviceConfig.serviceType,
        remoteServiceId: remoteParticipant.serviceId,
        remoteServiceDescription,
    });
}
function mapRoleConfigToDevices(pairwiseServiceConfigurations, experiment) {
    if (!experiment.devices || experiment.devices.length === 0) {
        throw new MissingPropertyError('Experiment must have a device to be run', 400);
    }
    const deviceMappedServiceConfigs = [];
    for (const serviceConfig of pairwiseServiceConfigurations) {
        if (!serviceConfig.participants || serviceConfig.participants.length !== 2) {
            throw new MissingPropertyError('pairwiseServiceConfigurations must have exactly two participants', 400);
        }
        const devicesA = experiment.devices
            .filter(d => d.role === serviceConfig.participants[0].role)
            .map(d => d);
        const devicesB = experiment.devices
            .filter(d => d.role === serviceConfig.participants[1].role)
            .map(d => d);
        for (const deviceA of devicesA) {
            for (const deviceB of devicesB) {
                if (deviceA !== deviceB) {
                    deviceMappedServiceConfigs.push({
                        ...serviceConfig,
                        devices: [deviceA, deviceB],
                    });
                }
                else {
                    // TODO: Handle same device
                }
            }
        }
    }
    return deviceMappedServiceConfigs;
}
function toPairwiseServiceConfig(serviceConfigurations) {
    const pairwiseServiceConfigurations = [];
    for (const serviceConfig of serviceConfigurations) {
        const participants = serviceConfig.participants;
        if (participants) {
            for (let i = 0; i < participants.length; i++) {
                for (let j = i + 1; j < participants.length; j++) {
                    const participantA = participants[i];
                    const participantB = participants[j];
                    // TODO: Check for Service Direction
                    pairwiseServiceConfigurations.push({
                        ...serviceConfig,
                        participants: [participantA, participantB],
                    });
                }
            }
        }
    }
    logger.log('info', 'Built pairwise service configurations', {
        data: { pairwiseServiceConfigurations },
    });
    return pairwiseServiceConfigurations;
}
async function createPeerconnection(type, serviceConfig) {
    switch (type) {
        case 'webrtc':
            return createPeerconnectionWebrtc(serviceConfig);
        case 'websocket':
            return await createPeerconnectionWebsocket(serviceConfig);
    }
}
function createPeerconnectionWebrtc(serviceConfig) {
    return {
        type: 'webrtc',
        devices: [
            {
                url: getUrlOrInstanceUrl(serviceConfig.devices[0]),
            },
            {
                url: getUrlOrInstanceUrl(serviceConfig.devices[1]),
            },
        ],
        configuration: {
            iceServers: [
                { urls: config.STUN_SERVER_URL },
                {
                    urls: config.TURN_SERVER_URL,
                    username: config.TURN_SERVER_USERNAME,
                    credential: config.TURN_SERVER_CREDENTIAL,
                },
            ],
        },
    };
}
async function createPeerconnectionWebsocket(serviceConfig) {
    const room = await clients.forwarding.createRoom({
        participants: [{ id: randomUUID() }, { id: randomUUID() }],
    });
    return {
        type: 'websocket',
        devices: [
            {
                url: getUrlOrInstanceUrl(serviceConfig.devices[0]),
            },
            {
                url: getUrlOrInstanceUrl(serviceConfig.devices[1]),
            },
        ],
        configuration: {
            webSocketUrls: [0, 1].map(index => `${room.url}?${new URLSearchParams({
                id: room?.participants[index].id,
            }).toString()}`
                .replace('http://', 'ws://')
                .replace('https://', 'wss://')),
        },
    };
}
function intersection(array1, array2) {
    return array1.filter(value => array2.includes(value));
}
function getResolvedDevice(resolvedDevices, url) {
    const resolvedDevice = resolvedDevices.find(device => device.url === url);
    if (!resolvedDevice) {
        throw new MissingEntityError(`Could not find a resolved device for device "${url}"`, 404);
    }
    return resolvedDevice;
}
function getSupportedConnectionTypes(device, serviceType, serviceId) {
    const supportedConnectionTypes = device.services?.find(service => service.serviceType === serviceType && service.serviceId === serviceId)?.supportedConnectionTypes;
    if (!supportedConnectionTypes)
        return ['webrtc'];
    return supportedConnectionTypes.filter(supportedConnectionType => supportedConnectionType === 'webrtc' || supportedConnectionType === 'websocket');
}
//# sourceMappingURL=connectionPlan.js.map